"""Data Handler using kz-filtered data."""

__author__ = 'Lukas Leufen'
__date__ = '2020-08-26'

import inspect
import numpy as np
import pandas as pd
import xarray as xr
from typing import List, Union

from mlair.data_handler.data_handler_single_station import DataHandlerSingleStation
from mlair.data_handler import DefaultDataHandler
from mlair.helpers import remove_items, to_list, TimeTrackingWrapper
from mlair.helpers.statistics import KolmogorovZurbenkoFilterMovingWindow as KZFilter

# define a more general date type for type hinting
str_or_list = Union[str, List[str]]


class DataHandlerKzFilterSingleStation(DataHandlerSingleStation):
    """Data handler for a single station to be used by a superior data handler. Inputs are kz filtered."""

    _requirements = remove_items(inspect.getfullargspec(DataHandlerSingleStation).args, ["self", "station"])

    def __init__(self, *args, kz_filter_length, kz_filter_iter, **kwargs):
        assert kwargs.get("sampling") == "hourly"  # This data handler requires hourly data resolution
        kz_filter_length = to_list(kz_filter_length)
        kz_filter_iter = to_list(kz_filter_iter)
        # self.original_data = None  # ToDo: implement here something to store unfiltered data
        self.kz_filter_length = kz_filter_length
        self.kz_filter_iter = kz_filter_iter
        self.cutoff_period = None
        self.cutoff_period_days = None
        super().__init__(*args, **kwargs)

    def setup_samples(self):
        """
        Setup samples. This method prepares and creates samples X, and labels Y.
        """
        self.load_data()
        self.interpolate(dim=self.time_dim, method=self.interpolation_method, limit=self.interpolation_limit)
        self.set_inputs_and_targets()
        self.apply_kz_filter()
        # this is just a code snippet to check the results of the kz filter
        # import matplotlib
        # matplotlib.use("TkAgg")
        # import matplotlib.pyplot as plt
        # self.input_data.data.sel(filter="74d", variables="temp", Stations="DEBW107").plot()
        # self.input_data.data.sel(variables="temp", Stations="DEBW107").plot.line(hue="filter")
        if self.do_transformation is True:
            self.call_transform()
        self.make_samples()

    @TimeTrackingWrapper
    def apply_kz_filter(self):
        """Apply kolmogorov zurbenko filter only on inputs."""
        kz = KZFilter(self.input_data.data, wl=self.kz_filter_length, itr=self.kz_filter_iter, filter_dim="datetime")
        filtered_data: List[xr.DataArray] = kz.run()
        self.cutoff_period = kz.period_null()
        self.cutoff_period_days = kz.period_null_days()
        self.input_data.data = xr.concat(filtered_data, pd.Index(self.create_filter_index(), name="filter"))

    def create_filter_index(self) -> pd.Index:
        """
        Round cut off periods in days and append 'res' for residuum index.

        Round small numbers (<10) to single decimal, and higher numbers to int. Transform as list of str and append
        'res' for residuum index.
        """
        index = np.round(self.cutoff_period_days, 1)
        f = lambda x: int(np.round(x)) if x >= 10 else np.round(x, 1)
        index = list(map(f, index.tolist()))
        index = list(map(lambda x: str(x) + "d", index)) + ["res"]
        return pd.Index(index, name="filter")

    def get_transposed_history(self) -> xr.DataArray:
        """Return history.

        :return: history with dimensions datetime, window, Stations, variables.
        """
        return self.history.transpose("datetime", "window", "Stations", "variables", "filter").copy()


class DataHandlerKzFilter(DefaultDataHandler):
    """Data handler using kz filtered data."""

    data_handler = DataHandlerKzFilterSingleStation
    data_handler_transformation = DataHandlerKzFilterSingleStation
    _requirements = data_handler.requirements()
