
__author__ = 'Lukas Leufen'
__date__ = '2020-07-07'

from collections import Iterator, Iterable
import keras
import numpy as np
import math
import os
import shutil
import pickle
from typing import Tuple, List


class StandardIterator(Iterator):

    _position: int = None

    def __init__(self, collection: list):
        assert isinstance(collection, list)
        self._collection = collection
        self._position = 0

    def __next__(self):
        """Return next element or stop iteration."""
        try:
            value = self._collection[self._position]
            self._position += 1
        except IndexError:
            raise StopIteration()
        return value


class DataCollection(Iterable):

    def __init__(self, collection: list = None):
        if collection is None:
            collection = []
        assert isinstance(collection, list)
        self._collection = collection
        self._mapping = {}
        self._set_mapping()

    def __len__(self):
        return len(self._collection)

    def __iter__(self) -> Iterator:
        return StandardIterator(self._collection)

    def __getitem__(self, index):
        if isinstance(index, int):
            return self._collection[index]
        else:
            return self._collection[self._mapping[str(index)]]

    def add(self, element):
        self._collection.append(element)
        self._mapping[str(element)] = len(self._collection)

    def _set_mapping(self):
        for i, e in enumerate(self._collection):
            self._mapping[str(e)] = i

    def keys(self):
        return list(self._mapping.keys())


class KerasIterator(keras.utils.Sequence):

    def __init__(self, collection: DataCollection, batch_size: int, batch_path: str, shuffle_batches: bool = False,
                 model=None, upsampling=False, name=None):
        self._collection = collection
        batch_path = os.path.join(batch_path, str(name if name is not None else id(self)))
        self._path = os.path.join(batch_path, "%i.pickle")
        self.batch_size = batch_size
        self.model = model
        self.shuffle = shuffle_batches
        self.upsampling = upsampling
        self.indexes: list = []
        self._cleanup_path(batch_path)
        self._prepare_batches()

    def __len__(self) -> int:
        return len(self.indexes)

    def __getitem__(self, index: int) -> Tuple[np.ndarray, np.ndarray]:
        """Get batch for given index."""
        return self.__data_generation(self.indexes[index])

    def _get_model_rank(self):
        if self.model is not None:
            mod_out = self.model.output_shape
            if isinstance(mod_out, tuple):  # only one output branch: (None, ahead)
                mod_rank = 1
            elif isinstance(mod_out, list):  # multiple output branches, e.g.: [(None, ahead), (None, ahead)]
                mod_rank = len(mod_out)
            else:  # pragma: no cover
                raise TypeError("model output shape must either be tuple or list.")
            return mod_rank
        else:  # no model provided, assume to use single output
            return 1

    def __data_generation(self, index: int) -> Tuple[np.ndarray, np.ndarray]:
        """Load pickle data from disk."""
        file = self._path % index
        with open(file, "rb") as f:
            data = pickle.load(f)
        return data["X"], data["Y"]

    @staticmethod
    def _concatenate(new: List[np.ndarray], old: List[np.ndarray]) -> List[np.ndarray]:
        """Concatenate two lists of data along axis=0."""
        return list(map(lambda n1, n2: np.concatenate((n1, n2), axis=0), old, new))

    def _get_batch(self, data_list: List[np.ndarray], b: int) -> List[np.ndarray]:
        """Get batch according to batch size from data list."""
        return list(map(lambda data: data[b * self.batch_size:(b+1) * self.batch_size, ...], data_list))

    def _permute_data(self, X, Y):
        p = np.random.permutation(len(X[0]))  # equiv to .shape[0]
        X = list(map(lambda x: x[p], X))
        Y = list(map(lambda x: x[p], Y))
        return X, Y

    def _prepare_batches(self) -> None:
        """
        Prepare all batches as locally stored files.

        Walk through all elements of collection and split (or merge) data according to the batch size. Too long data
        sets are divided into multiple batches. Not fully filled batches are merged with data from the next collection
        element. If data is remaining after the last element, it is saved as smaller batch. All batches are enumerated
        beginning from 0. A list with all batch numbers is stored in class's parameter indexes.
        """
        index = 0
        remaining = None
        mod_rank = self._get_model_rank()
        for data in self._collection:
            X = data.get_X(upsampling=self.upsampling)
            Y = [data.get_Y(upsampling=self.upsampling)[0] for _ in range(mod_rank)]
            if self.upsampling:
                X, Y = self._permute_data(X, Y)
            if remaining is not None:
                X, Y = self._concatenate(X, remaining[0]), self._concatenate(Y, remaining[1])
            length = X[0].shape[0]
            batches = self._get_number_of_mini_batches(length)
            for b in range(batches):
                batch_X, batch_Y = self._get_batch(X, b), self._get_batch(Y, b)
                self._save_to_pickle(X=batch_X, Y=batch_Y, index=index)
                index += 1
            if (batches * self.batch_size) < length:  # keep remaining to concatenate with next data element
                remaining = (self._get_batch(X, batches), self._get_batch(Y, batches))
            else:
                remaining = None
        if remaining is not None:  # add remaining as smaller batch
            self._save_to_pickle(X=remaining[0], Y=remaining[1], index=index)
            index += 1
        self.indexes = np.arange(0, index).tolist()

    def _save_to_pickle(self, X: List[np.ndarray], Y: List[np.ndarray], index: int) -> None:
        """Save data as pickle file with variables X and Y and given index as <index>.pickle ."""
        data = {"X": X, "Y": Y}
        file = self._path % index
        with open(file, "wb") as f:
            pickle.dump(data, f)

    def _get_number_of_mini_batches(self, number_of_samples: int) -> int:
        """Return number of mini batches as the floored ration of number of samples to batch size."""
        return math.floor(number_of_samples / self.batch_size)

    @staticmethod
    def _cleanup_path(path: str, create_new: bool = True) -> None:
        """First remove existing path, second create empty path if enabled."""
        if os.path.exists(path):
            shutil.rmtree(path)
        if create_new is True:
            os.makedirs(path)

    def on_epoch_end(self) -> None:
        """Randomly shuffle indexes if enabled."""
        if self.shuffle is True:
            np.random.shuffle(self.indexes)


class DummyData:  # pragma: no cover

    def __init__(self, number_of_samples=np.random.randint(100, 150)):
        self.number_of_samples = number_of_samples

    def get_X(self):
        X1 = np.random.randint(0, 10, size=(self.number_of_samples, 14, 5))  # samples, window, variables
        X2 = np.random.randint(21, 30, size=(self.number_of_samples, 10, 2))  # samples, window, variables
        X3 = np.random.randint(-5, 0, size=(self.number_of_samples, 1, 2))  # samples, window, variables
        return [X1, X2, X3]

    def get_Y(self):
        Y1 = np.random.randint(0, 10, size=(self.number_of_samples, 5, 1))  # samples, window, variables
        Y2 = np.random.randint(21, 30, size=(self.number_of_samples, 5, 1))  # samples, window, variables
        return [Y1, Y2]


if __name__ == "__main__":

    collection = []
    for _ in range(3):
        collection.append(DummyData(50))

    data_collection = DataCollection(collection=collection)

    path = os.path.join(os.path.dirname(os.path.abspath(__file__)), "testdata")
    iterator = KerasIterator(data_collection, 25, path, shuffle=True)

    for data in data_collection:
        print(data)